##
# File: ExeceUtils.py
# Date: 24-Jan-2020 jdw
#
##
import logging
import os
import subprocess
import sys

logger = logging.getLogger(__name__)


class ExecUtils(object):
    """Wrapper for subprocess execution."""

    def __init__(self):
        """Wrapper for subprocess execution"""

    def run(self, execPath, execArgList=None, outPath=None, outAppend=False, timeOut=None, inpPath=None, suppressStderr=False):
        """Execute the input program as a blocking subprocess with optional timeout.

        Args:
            execPath (str): path to executable program or script
            execArgList (list, optional): argument list. Defaults to None.
            outPath (str, optional): redirect stdout and stderr to this file handle. Defaults to None.
            inpPath (str, optional): redirect stdin to this file handle. Defaults to None.
            outAppend (bool, optional): append output. Defaults to False.
            timeOut (float, optional): timeout (seconds). Defaults to None.
            suppressStderr (bool, optional): suppress stderr output (default: combined with stdout)


        Returns:
            bool: true for sucess or False otherwise
        """
        retCode = False
        subProcResult = None
        kwD = {}
        if suppressStderr:
            myStderr = subprocess.DEVNULL
        else:
            myStderr = subprocess.STDOUT
        try:
            if not os.path.isfile(execPath) and os.access(execPath, os.X_OK):
                return retCode
            if sys.version_info[0] > 2 and timeOut:
                kwD = {"timeout": timeOut}
            cmdL = [execPath]
            if execArgList:
                cmdL.extend(execArgList)
            if outPath and inpPath:
                myMode = "a" if outAppend else "w"
                with open(outPath, myMode) as ofh, open(inpPath, "r") as ifh:
                    subProcResult = subprocess.call(cmdL, stdout=ofh, stdin=ifh, stderr=myStderr, **kwD)
            elif outPath:
                myMode = "a" if outAppend else "w"
                with open(outPath, myMode) as ofh:
                    subProcResult = subprocess.call(cmdL, stdout=ofh, stderr=myStderr, **kwD)
            else:
                subProcResult = subprocess.call(cmdL, **kwD)
            retCode = not subProcResult
        except Exception as e:
            logger.exception("Failing execution of %s (%s) with %s", execPath, execArgList, str(e))
        #
        if subProcResult != 0:
            logger.error("return code is %r", subProcResult)
        return retCode

    def runShell(self, script, outPath=None, outAppend=False, timeOut=None, inpPath=None, suppressStderr=False):
        """Execute the input program as a blocking subprocess with optional timeout.

        Args:
            script (str): script
            outPath (str, optional): redirect stdout and stderr to this file handle. Defaults to None.
            inpPath (str, optional): redirect stdin to this file handle. Defaults to None.
            outAppend (bool, optional): append output. Defaults to False.
            timeOut (float, optional): timeout (seconds). Defaults to None.
            suppressStderr (bool, optional): suppress stderr output (default: combined with stdout)


        Returns:
            bool: true for sucess or False otherwise
        """
        retCode = 0
        kwD = {}
        subProcResult = None
        if suppressStderr:
            myStderr = subprocess.DEVNULL
        else:
            myStderr = subprocess.STDOUT
        try:
            if sys.version_info[0] > 2 and timeOut:
                kwD = {"timeout": timeOut}
            #
            if outPath and inpPath:
                myMode = "a" if outAppend else "w"
                with open(outPath, myMode) as ofh, open(inpPath, "r") as ifh:
                    subProcResult = subprocess.run(script, stdout=ofh, stdin=ifh, stderr=myStderr, shell=True, check=False, **kwD)
            elif outPath:
                myMode = "a" if outAppend else "w"
                with open(outPath, myMode) as ofh:
                    subProcResult = subprocess.run(script, stdout=ofh, stderr=myStderr, shell=True, check=False, **kwD)
            else:
                subProcResult = subprocess.run(script, shell=True, check=False, **kwD)
            retCode = not subProcResult.returncode
        except Exception as e:
            logger.exception("Failing execution of %r with %s", subProcResult, str(e))
        #
        if subProcResult and subProcResult.returncode != 0:
            logger.warning("return code is %r", subProcResult.returncode)
        return retCode



##
# File: StashUtil.py
#
# Utilities to stash and recover a data in collection of directories to and from
# remote sftp, http or local POSIX file storage resources.
#
# Updates:
#
##

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging
import os

from rcsb.utils.io.FileUtil import FileUtil
from rcsb.utils.io.SftpUtil import SftpUtil

logger = logging.getLogger(__name__)


class StashUtil(object):
    """ Utilities to stash and recover a data in collection of (sub)directories to/from
        remote sftp, http or local POSIX file storage resources.
    """

    def __init__(self, localBundlePath, baseBundleFileName):
        """Set the bundle file name and path for class instance.

        Args:
            localBundlePath (str): writeable local path for bundle file artifact
            baseBundleFileName (str): bundle file name without file extension
        """
        #
        fn = baseBundleFileName
        self.__baseBundleFileName = fn + ".tar.gz"
        self.__localStashTarFilePath = os.path.join(localBundlePath, self.__baseBundleFileName)

    def makeBundle(self, localParentPath, subDirList):
        """ Bundle the subdirectories of the input parent directory path.

        Args:
            localParentPath (str): local parent directory path containing the bundling targets
            subDirList (list, str): list of subdirectories of the parent path to be bundled

        Returns:
            (bool): True for success or False otherwise
        """
        fileU = FileUtil()
        dirPathList = [os.path.join(localParentPath, subDir) for subDir in subDirList]
        okT = fileU.bundleTarfile(self.__localStashTarFilePath, dirPathList, mode="w:gz", recursive=True)
        return okT

    def storeBundle(self, url, remoteDirPath, remoteStashPrefix="A", userName=None, password=None):
        """ Store a copy of the bundled search dependencies remotely -

        Args:
            url (str): URL string for the destination host (e.g. sftp://myserver.net or None for a local file)
            remoteDirPath (str): remote directory path on the remote resource
            remoteStashPrefix (str, optional): optional label preppended to the stashed dependency bundle artifact (default='A')
            userName (str, optional): optional access information. Defaults to None.
            password (str, optional): optional access information. Defaults to None.

        Returns:
          bool:  True for success or False otherwise

        """
        try:
            ok = False
            fn = self.__makeBundleFileName(self.__baseBundleFileName, remoteStashPrefix=remoteStashPrefix)
            if url and url.startswith("sftp://"):
                sftpU = SftpUtil()
                hostName = url[7:]
                ok = sftpU.connect(hostName, userName, pw=password, port=22)
                if ok:
                    remotePath = os.path.join("/", remoteDirPath, fn)
                    ok = sftpU.put(self.__localStashTarFilePath, remotePath)
            elif not url:
                fileU = FileUtil()
                remotePath = os.path.join(remoteDirPath, fn)
                ok = fileU.put(self.__localStashTarFilePath, remotePath)
            else:
                logger.error("Unsupported stash protocol %r", url)
            return ok
        except Exception as e:
            logger.exception("For %r %r failing with %s", url, remoteDirPath, str(e))
        return False

    def fetchBundle(self, localRestoreDirPath, url, remoteDirPath, remoteStashPrefix="A", userName=None, password=None):
        """Restore bundled dependencies from remote storage and unbundle these in the
           current local cache directory.

        Args:
            localRestoreDirPath (str): local restore path
            url (str): remote URL
            remoteDirPath (str): remote directory path on the remote resource
            remoteStashPrefix (str, optional): optional label preppended to the stashed dependency bundle artifact (default='A')
            userName (str, optional): optional access information. Defaults to None.
            password (str, optional): optional access information. Defaults to None.
        """
        try:
            ok = False
            fileU = FileUtil()
            fn = self.__makeBundleFileName(self.__baseBundleFileName, remoteStashPrefix=remoteStashPrefix)
            if not url:
                remotePath = os.path.join(remoteDirPath, fn)
                ok = fileU.get(remotePath, self.__localStashTarFilePath)

            elif url and (url.startswith("http://") or url.startswith("https://")):
                remotePath = url + os.path.join("/", remoteDirPath, fn)
                ok = fileU.get(remotePath, self.__localStashTarFilePath)

            elif url and url.startswith("sftp://"):
                sftpU = SftpUtil()
                ok = sftpU.connect(url[7:], userName, pw=password, port=22)
                if ok:
                    remotePath = os.path.join(remoteDirPath, fn)
                    ok = sftpU.get(remotePath, self.__localStashTarFilePath)
            else:
                logger.error("Unsupported protocol %r", url)
            if ok:
                ok = fileU.unbundleTarfile(self.__localStashTarFilePath, dirPath=localRestoreDirPath)
            return ok
        except Exception as e:
            logger.exception("For %r %r Failing with %s", url, remoteDirPath, str(e))
            ok = False
        return ok

    def __makeBundleFileName(self, baseBundleFileName, remoteStashPrefix="A"):
        fn = baseBundleFileName
        try:
            fn = "%s-%s" % (remoteStashPrefix.upper(), baseBundleFileName) if remoteStashPrefix else baseBundleFileName
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return fn



##
# File: UrlRequestUtil.py
#
# Simple wrapper for URL request processing
#
# Updates:
#  17-Mar-2019 jdw adjust return value to include response error code
#  16-Dec-2019 jdw add HTTPException to retry()
#  28-May-2029 jdw unwrapped methods now using requests module library.
#
##

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"


import contextlib
import json
import logging
import ssl
import warnings

import requests


from requests.adapters import HTTPAdapter
from urllib3.util import Retry
from rcsb.utils.io.decorators import retry


try:
    from http.client import HTTPException
    from urllib.parse import urlencode
    from urllib.request import urlopen, Request, URLError, HTTPError

    # from http.client import HTTPException, RemoteDisconnected
except ImportError:
    # pylint: disable=ungrouped-imports
    from httplib import HTTPException
    from urllib import urlencode
    from urllib2 import urlopen, Request, URLError, HTTPError

logger = logging.getLogger(__name__)
warnings.filterwarnings("ignore", message="Unverified HTTPS request")


class UrlRequestUtil(object):
    """Simple wrapper for URL request processing"""

    def __init__(self, **kwargs):
        pass

    def postUnWrapped(self, url, endPoint, paramD, **kwargs):
        return self.__postRequests(url, endPoint, paramD, **kwargs)

    @retry((URLError, HTTPError, HTTPException), maxAttempts=3, delaySeconds=5, multiplier=3, defaultValue=(None, None), logger=logger)
    def post(self, url, endPoint, paramD, **kwargs):
        return self.__post(url, endPoint, paramD, **kwargs)

    def __post(self, url, endPoint, paramD, **kwargs):
        """"""
        ret = None
        retCode = None
        headerL = kwargs.get("headers", [])
        sslCert = kwargs.get("sslCert", "disable")
        exceptionsCatch = kwargs.get("exceptionsCatch", (HTTPError))
        httpCodesCatch = kwargs.get("httpCodesCatch", [])
        encoding = kwargs.get("encoding", "utf-8")
        #
        returnContentType = kwargs.get("returnContentType", None)
        if returnContentType == "JSON":
            if ("Accept", "application/json") not in headerL:
                headerL.append(("Accept", "application/json"))
        #
        optD = {"timeout": 5}
        try:
            if sslCert == "disable":
                ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
                ctx.check_hostname = False
                ctx.verify_mode = ssl.CERT_NONE
                ctx.options &= ~ssl.OP_NO_SSLv3  # this does not seem to work has expected
                # ctx = ssl._create_unverified_context()  # pylint: disable=protected-access
                optD = {"context": ctx}
            logger.debug("Support for SSLv3 %r TLSv1_1 %r TLSv1_2 %r", ssl.HAS_SSLv3, ssl.HAS_TLSv1_1, ssl.HAS_TLSv1_2)
            #
            urlPath = "%s/%s" % (url, endPoint)
            requestData = urlencode(paramD).encode(encoding)
            logger.debug("Request %s with data %r", urlPath, requestData)

            with contextlib.closing(urlopen(urlPath, requestData, **optD)) as req:
                # pylint: disable=no-member
                #
                # for hTup in headerL:
                #    req.add_header(hTup[0], hTup[1])
                ret = req.read()
                retCode = req.getcode()
        #
        except exceptionsCatch as e:
            logger.debug("status code %r", e.code)
            retCode = e.code
            if retCode not in httpCodesCatch:
                raise e
        except Exception as e:
            logger.error("Failing %s %s %r with %s", url, endPoint, paramD, str(e))
            raise e
        #
        try:
            if returnContentType == "JSON":
                return json.loads(ret.decode(encoding)), retCode
            else:
                return ret.decode(encoding), retCode
        except Exception as e:
            if ret:
                logger.error("Decode failing %s %s %r with %s", url, endPoint, paramD, str(e))
                logger.debug("End of return is %r", ret[-1000:])

        return None, retCode

    def getUnWrapped(self, url, endPoint, paramD, **kwargs):
        return self.__getRequests(url, endPoint, paramD, **kwargs)

    @retry((URLError, HTTPError, HTTPException), maxAttempts=3, delaySeconds=5, multiplier=3, defaultValue=(None, None), logger=logger)
    def get(self, url, endPoint, paramD, **kwargs):
        return self.__get(url, endPoint, paramD, **kwargs)

    def __get(self, url, endPoint, paramD, **kwargs):
        """"""
        ret = None
        retCode = None
        sslCert = kwargs.get("sslCert", "disable")
        encoding = kwargs.get("encoding", "utf-8")
        headerL = kwargs.get("headers", [])
        exceptionsCatch = kwargs.get("exceptionsCatch", (HTTPError))
        httpCodesCatch = kwargs.get("httpCodesCatch", [])
        returnContentType = kwargs.get("returnContentType", None)
        if returnContentType == "JSON":
            if ("Accept", "application/json") not in headerL:
                headerL.append(("Accept", "application/json"))
        #
        headerL.append(("User-Agent", "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.95 Safari/537.36"))
        #
        optD = {"timeout": 5}
        try:
            if sslCert == "disable":
                ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
                ctx.check_hostname = False
                ctx.verify_mode = ssl.CERT_NONE
                ctx.options &= ~ssl.OP_NO_SSLv3  # this does not seem to work has expected
                # ctx = ssl._create_unverified_context()  # pylint: disable=protected-access
                optD = {"context": ctx}
            logger.debug("Support for SSLv3 %r TLSv1_1 %r TLSv1_2 %r", ssl.HAS_SSLv3, ssl.HAS_TLSv1_1, ssl.HAS_TLSv1_2)

            #
            urlPath = "%s/%s" % (url, endPoint)
            requestData = urlencode(paramD)
            logger.debug("Request %s with parameters %s", urlPath, requestData)

            req = Request("%s?%s" % (urlPath, requestData))
            if headerL:
                for hTup in headerL:
                    req.add_header(hTup[0], hTup[1])
            #
            with contextlib.closing(urlopen(req, **optD)) as req:
                # pylint: disable=no-member
                ret = req.read()
                retCode = req.getcode()
            #
            #    req = urlopen(req, **optD)
            #    ret = req.read()
            #    retCode = req.getcode()

            if ret is None:
                logger.error("Return is empty - return code is %r", retCode)
            else:
                logger.debug("Return length %d return code %r", len(ret), retCode)
            #
        except exceptionsCatch as e:
            retCode = e.code
            if retCode not in httpCodesCatch:
                raise e
        except Exception as e:
            logger.error("Failing %s %s %r with %s", url, endPoint, paramD, str(e))
            raise e

        try:
            if returnContentType == "JSON":
                return json.loads(ret.decode(encoding)), retCode
            else:
                return ret.decode(encoding), retCode
        except Exception as e:
            if ret:
                logger.error("Decode failing %s %s %r with %s", url, endPoint, paramD, str(e))
                logger.debug("End of return is %r", ret[-1000:])

        return None, retCode

    def __getRequests(self, url, endPoint, paramD, **kwargs):
        """"""
        ret = None
        retCode = None
        sslCert = kwargs.get("sslCert", "disable")
        verify = False if sslCert == "disable" else True

        # encoding = kwargs.get("encoding", "utf-8")
        headerD = kwargs.get("headers", {})
        exceptionsCatch = kwargs.get("exceptionsCatch", (HTTPError))
        httpCodesCatch = kwargs.get("httpCodesCatch", [])
        returnContentType = kwargs.get("returnContentType", None)
        timeOutSeconds = kwargs.get("timeOut", 5)
        retries = kwargs.get("retries", 3)
        statusForcelist = (429, 500, 502, 503, 504)
        backoffFactor = 5
        methodWhitelist = ("HEAD", "GET", "PUT", "DELETE", "OPTIONS", "TRACE", "POST")
        if returnContentType == "JSON":
            if "Accept" not in headerD:
                headerD["Accept"] = "application/json"
        #
        headerD.update({"User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.95 Safari/537.36"})
        #
        optD = {"timeout": timeOutSeconds, "allow_redirects": True, "verify": verify}
        try:
            #
            urlPath = "%s/%s" % (url, endPoint)
            if retries:
                with requests.Session() as session:
                    thisRetry = Retry(
                        total=retries,
                        read=retries,
                        connect=retries,
                        backoff_factor=backoffFactor,
                        status_forcelist=statusForcelist,
                        # allowed_methods=methodWhitelist,
                        method_whitelist=methodWhitelist,
                    )
                    adapter = HTTPAdapter(max_retries=thisRetry)
                    session.mount("http://", adapter)
                    session.mount("https://", adapter)
                    # session = self.__requestsRetrySession(retries=3, backoffFactor=5, statusForcelist=(429, 500, 502, 503, 504))
                    req = session.get(urlPath, params=paramD, headers=headerD, **optD)
            else:
                req = requests.get(urlPath, params=paramD, headers=headerD, **optD)
            retCode = req.status_code
            if retCode == 200:
                if returnContentType == "JSON":
                    ret = req.json()
                else:
                    ret = req.text
            #
            if ret is None:
                logger.debug("Return is empty - return code is %r", retCode)
            else:
                logger.debug("Return length %d return code %r", len(ret), retCode)
            #
            return ret, retCode
        except exceptionsCatch as e:
            retCode = e.code
            if retCode not in httpCodesCatch:
                raise e
        except Exception as e:
            logger.error("Failing %s %s %r with %s", url, endPoint, paramD, str(e))
            raise e

        return None, retCode

    def __postRequests(self, url, endPoint, paramD, **kwargs):
        """"""
        ret = None
        retCode = None
        sslCert = kwargs.get("sslCert", "disable")
        verify = False if sslCert == "disable" else True
        # encoding = kwargs.get("encoding", "utf-8")
        headerD = kwargs.get("headers", {})
        exceptionsCatch = kwargs.get("exceptionsCatch", (HTTPError))
        httpCodesCatch = kwargs.get("httpCodesCatch", [])
        returnContentType = kwargs.get("returnContentType", None)
        timeOutSeconds = kwargs.get("timeOut", 5)
        retries = kwargs.get("retries", 3)
        statusForcelist = (429, 500, 502, 503, 504)
        methodWhitelist = ("HEAD", "GET", "PUT", "DELETE", "OPTIONS", "TRACE", "POST")
        backoffFactor = 5
        if returnContentType == "JSON":
            if "Accept" not in headerD:
                headerD["Accept"] = "application/json"
        #
        headerD.update({"User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.95 Safari/537.36"})
        #
        optD = {"timeout": timeOutSeconds, "allow_redirects": True, "verify": verify}
        try:
            urlPath = "%s/%s" % (url, endPoint)
            if retries:
                with requests.Session() as session:
                    thisRetry = Retry(
                        total=retries,
                        read=retries,
                        connect=retries,
                        backoff_factor=backoffFactor,
                        status_forcelist=statusForcelist,
                        # allowed_methods=methodWhitelist,
                        method_whitelist=methodWhitelist,
                    )
                    adapter = HTTPAdapter(max_retries=thisRetry)
                    session.mount("http://", adapter)
                    session.mount("https://", adapter)
                    # session = self.__requestsRetrySession(retries=retries, backoffFactor=5, statusForcelist=(429, 500, 502, 504))
                    req = session.post(urlPath, data=paramD, headers=headerD, **optD)
            else:
                req = requests.post(urlPath, data=paramD, headers=headerD, **optD)
            retCode = req.status_code
            if retCode == 200:
                if returnContentType == "JSON":
                    ret = req.json()
                else:
                    ret = req.text
            #
            if ret is None:
                logger.debug("Return is empty - return code is %r", retCode)
            else:
                logger.debug("Return length %d return code %r", len(ret), retCode)
            #
            return ret, retCode
        except exceptionsCatch as e:
            retCode = e.code
            if retCode not in httpCodesCatch:
                raise e
        except Exception as e:
            logger.error("Failing %s %s %r with %s", url, endPoint, paramD, str(e))
            raise e

        return None, retCode



##
# File: FileUtil.py
#
# Skeleton implementation for File I/O
#
# Updates:
#  5-Jun-2018 jdw add support for local copy operations using shutil.copy
#  6-Mar-2019 jdw add previously stubbed remote access and tar file member support
# 10-Mar-2019 jdw add exists() method
# 11-Aug-2019 jdw incorporate compress/uncompreess and toAscii methods from IoUtil.py
#                 add __extractZipMember() method.  Add retry support for __fetchUrl().
#                 Replace urlibx version of __fetchUrl with a version that depends on
#                 requests module to better support redirection and authentication.
##

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

#
import bz2
import contextlib
import gzip
import io
import logging
import os
import shutil
import tarfile
import tempfile
import zipfile

import requests
from rcsb.utils.io.decorators import retry

# pylint: disable=ungrouped-imports
try:
    import urllib.request as myurl
except Exception:
    import urllib2 as myurl

try:
    from urllib.parse import urlsplit
except Exception:
    from urlparse import urlsplit


logger = logging.getLogger(__name__)


class FileUtil(object):
    """Skeleton implementation for File I/O operations"""

    def __init__(self, workPath=None, **kwargs):
        _ = kwargs
        self.__workPath = workPath
        if self.__workPath and self.__workPath != ".":
            self.mkdir(workPath)

    def exists(self, filePath, mode=os.R_OK):
        """Return if the input file exists with optional mode settings

        Args:
            filePath (string): local file path
            mode (mode, optional): mode setting. Defaults to os.R_OK.

        Returns:
            bool: True if file exists or False otherwise
        """
        try:
            return os.access(filePath, mode)
        except Exception:
            pass
        return False

    def mkdir(self, dirPath, mode=0o755):
        """Create the input directory.

        Args:
            dirPath (string): local directory path
            mode (mode, optional): mode setting. Defaults to os.W_OK.

        Returns:
            bool: True for success or False otherwise
        """
        try:
            logger.debug("Checking target directory %s", dirPath)
            if not os.access(dirPath, os.W_OK):
                logger.debug("Creating cache directory %s", dirPath)
                os.makedirs(dirPath, mode)
            return True
        except Exception as e:
            logger.exception("Failing for %s with %s", dirPath, str(e))
        return False

    def mkdirForFile(self, filePath, mode=0o755):
        """Create directory path for the input path if it does not exists.

        Args:
            filePath (string): local file path target
            mode (mode, optional): mode setting. Defaults to os.W_OK.

        Returns:
            bool: True for success or False otherwise
        """
        try:
            logger.debug("Checking target directory for file %s", filePath)
            dirPath, _ = os.path.split(filePath)
            if dirPath:
                return self.mkdir(dirPath, mode)
            else:
                return True
        except Exception as e:
            logger.exception("Failing for %s with %s", filePath, str(e))
        return False

    #
    def remove(self, pth):
        """Method to remove input file, link  or directory."""
        ok = False
        try:
            if pth and not self.exists(pth):
                ok = True
            if os.path.isfile(pth) or os.path.islink(pth):
                os.unlink(pth)
                ok = True
            elif os.path.isdir(pth):
                shutil.rmtree(pth, True)
                ok = True
        except Exception as e:
            logger.error("Failing for %s with %s", pth, str(e))
            ok = False
        #
        return ok

    #
    def get(self, remote, local, **kwargs):
        """Fetch remote file to a local path.

        Arguments:
            remote (TYPE): source locator (remote)
            local (TYPE): destination locator (local)
            **kwargs: Description

                tarMember(str) : target member name in a remote tarfile

        Returns:
            bool: True for success or false otherwise

        """
        try:
            ret = False
            tarMember = kwargs.get("tarMember", None)
            #
            localFlag = self.isLocal(remote)
            #
            if localFlag and not tarMember:
                rPath = self.getFilePath(remote)
                lPath = self.getFilePath(local)
                self.mkdirForFile(lPath)
                shutil.copyfile(rPath, lPath)
                ret = True
            elif localFlag and tarMember:
                # Extract a particular member from a local tar file -
                rPath = self.getFilePath(remote)
                lPath = self.getFilePath(local)
                self.mkdirForFile(lPath)
                ret = self.__extractTarMember(rPath, tarMember, lPath)
                logger.debug("Extract %r from %r to %r status %r", tarMember, rPath, lPath, ret)
            elif not localFlag and tarMember:
                # Extract a particular member from a remote tar file
                tarPath = os.path.join(self.__workPath, self.getFileName(remote))
                ret = self.__fetchUrl(remote, tarPath, **kwargs)
                logger.debug("Fetched %r to %r status %r", remote, tarPath, ret)
                #
                ret = self.__extractTarMember(tarPath, tarMember, self.getFilePath(local)) if ret else False
                logger.debug("Extract %r from %r to %r status %r", tarMember, tarPath, self.getFilePath(local), ret)
            elif not localFlag and not tarMember:
                ret = self.__fetchUrl(remote, self.getFilePath(local), **kwargs)
                logger.debug("Fetched %r localpath %r status %r", remote, self.getFilePath(local), ret)
            else:
                ret = False
            #
        except Exception as e:
            logger.exception("For remote %r local %r failing with %s", remote, local, str(e))

        #
        return ret

    def put(self, local, remote, **kwargs):
        """Copy local file to remote location."""
        _ = kwargs
        try:
            ret = False
            localFlag = self.isLocal(remote)
            #
            if localFlag:
                rPath = self.getFilePath(remote)
                self.mkdirForFile(rPath)
                lPath = self.getFilePath(local)
                shutil.copyfile(lPath, rPath)
                ret = True
            else:
                ret = False
        except Exception as e:
            logger.exception("For remote %r local %r failing with %s", remote, local, str(e))
            ret = False
        return ret

    def getFilePath(self, locator):
        try:
            locSp = urlsplit(locator)
            return locSp.path
        except Exception as e:
            logger.exception("For locator %r failing with %s", locator, str(e))
        return None

    def getFileName(self, locator):
        try:
            locSp = urlsplit(locator)
            (_, fn) = os.path.split(locSp.path)
            return fn
        except Exception as e:
            logger.exception("For locator %r failing with %s", locator, str(e))
        return None

    def isLocal(self, locator):
        try:
            locSp = urlsplit(locator)
            return locSp.scheme in ["", "file"]
        except Exception as e:
            logger.exception("For locator %r failing with %s", locator, str(e))
        return None

    def getScheme(self, locator):
        try:
            locSp = urlsplit(locator)
            return locSp.scheme
        except Exception as e:
            logger.exception("For locator %r Failing with %s", locator, str(e))
        return None

    def bundleTarfile(self, tarFilePath, dirPathList, mode="w:gz", recursive=True):
        """Create a tar file bundle of the contents of the input directory path.

        Args:
            tarFilePath (str): output tar file path
            dirPathList (list): target directory path list to store
            mode (str, optional): the file mode for the tar file. Defaults to "w:gz".
            recursive (bool, optional): include subdirectories recursively. Defaults to True.

        Returns:
            bool: True for success or False otherwise
        """
        ret = True
        curDir = os.getcwd()
        try:
            self.mkdirForFile(tarFilePath)
            with tarfile.open(tarFilePath, mode=mode) as archive:
                for dirPath in dirPathList:
                    dp, tp = os.path.split(dirPath)
                    os.chdir(dp)
                    archive.add(tp, recursive=recursive)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
            ret = False
        os.chdir(curDir)
        return ret

    def unbundleTarfile(self, tarFilePath, dirPath="."):
        """Unbundle input tar file.

        Args:
            tarFilePath (str): path to the input tar bundle file
            dirPath (str, optional): directory path to write extracted data. Defaults to ".".

        Returns:
            bool: True for success or False otherwise
        """
        #   import tarfile contents into dirPath -
        ret = True
        try:

            with tarfile.open(tarFilePath) as tar:
                tar.extractall(path=dirPath)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
            ret = False
        return ret

    def unbundleZipfile(self, zipFilePath, dirPath="."):
        """Unbundle input ZIP file.

        Args:
            zipFilePath (str): path to the input tar bundle file
            dirPath (str, optional): directory path to write extracted data. Defaults to ".".

        Returns:
            bool: True for success or False otherwise
        """
        #   import zipfile contents into dirPath -
        ret = True
        try:
            with zipfile.ZipFile(zipFilePath, mode="r") as zObj:
                zObj.extractall(path=dirPath)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
            ret = False
        return ret

    def __extractTarMember(self, tarFilePath, memberName, memberPath):
        ret = True
        try:
            with tarfile.open(tarFilePath) as tar:
                fIn = tar.extractfile(memberName)
                with open(memberPath, "wb") as ofh:
                    ofh.write(fIn.read())
        except Exception as e:
            logger.exception("Failing with %s", str(e))
            ret = False
        return ret

    def __extractZipMember(self, zipFilePath, memberName, memberPath):
        ret = False
        try:
            with zipfile.ZipFile(zipFilePath, mode="r") as zObj:
                memberList = zObj.namelist()
                for member in memberList:
                    if member == memberName:
                        zObj.extract(member, path=memberPath)
                        ret = True
                        break
        except Exception as e:
            logger.exception("Failing with %s", str(e))
            ret = False
        return ret

    def __fetchUrl(self, url, filePath, **kwargs):

        try:
            noRetry = kwargs.get("noRetry", False)
            scheme = self.getScheme(url)
            ok = self.mkdirForFile(filePath)
            if not ok:
                logger.error("Failing to create target directory for file %r", filePath)
                return False
            if scheme in ["ftp"]:
                return self.__fetchUrlPy(url, filePath, **kwargs)
            else:
                if noRetry:
                    return self.__fetchUrlReqNoRt(url, filePath, **kwargs)
                else:
                    return self.__fetchUrlReqRt(url, filePath, **kwargs)
        except Exception as e:
            logger.error("Failing for url %r with %s", url, str(e))
        return False

    @retry((myurl.URLError, myurl.HTTPError), maxAttempts=3, delaySeconds=5, multiplier=3, defaultValue=False, logger=logger)
    def __fetchUrlPy(self, url, filePath, **kwargs):
        """Fetch data from a remote URL and store this in input filePath.

        Args:
            url (str): target URL to fetch
            filePath (str): path to store the data from the remote url
            username (str): basic auth username
            password (str): dasic auth password
            kwargs (dict, optional): other options

        Raises:
            e: any exception

        Returns:
            bool: True for sucess or False otherwise
        """
        try:
            user = kwargs.get("username", None)
            pw = kwargs.get("password", None)
            if user and pw:
                pwMgr = myurl.HTTPPasswordMgrWithDefaultRealm()
                pwMgr.add_password(None, url, user, pw)
                handler = myurl.HTTPBasicAuthHandler(pwMgr)
                opener = myurl.build_opener(handler)
                # Install the opener.
                myurl.install_opener(opener)
            #
            ok = self.mkdirForFile(filePath)
            if not ok:
                logger.error("Failing to create target directory for file %r", filePath)
                return False
            with open(filePath, "wb") as outFile:
                with contextlib.closing(myurl.urlopen(url)) as fp:
                    blockSize = 1024 * 8
                    while True:
                        block = fp.read(blockSize)  # pylint: disable=no-member
                        if not block:
                            break
                        outFile.write(block)
            return True
        except Exception as e:
            logger.error("Failing for %s with %s", filePath, str(e))
            raise e

        return False

    @retry((requests.exceptions.RequestException), maxAttempts=3, delaySeconds=5, multiplier=3, defaultValue=False, logger=logger)
    def __fetchUrlReqRaw(self, url, filePath, **kwargs):
        """Fetch data from a remote URL and store this in input filePath.

        Args:
            url (str): target URL to fetch
            filePath (str): path to store the data from the remote url
            username (str): basic auth username
            password (str): dasic auth password
            kwargs (dict, optional): other options

        Raises:
            e: any exception

        Returns:
            bool: True for sucess or False otherwise

            with open(filename, 'wb') as fd:
                for chunk in r.iter_content(chunk_size=128):
                    fd.write(chunk)


        """
        user = kwargs.get("username", None)
        pw = kwargs.get("password", None)
        #
        try:
            ok = self.mkdirForFile(filePath)
            if not ok:
                logger.error("Failing to create target directory for file %r", filePath)
                return False
            if user and pw:
                with requests.get(url, stream=True, auth=(user, pw), allow_redirects=True) as rIn:
                    with open(filePath, "wb") as fOut:
                        shutil.copyfileobj(rIn.raw, fOut)

            else:
                with requests.get(url, stream=True, allow_redirects=True) as rIn:
                    with open(filePath, "wb") as fOut:
                        shutil.copyfileobj(rIn.raw, fOut)

            return True
        except Exception as e:
            logger.error("Failing for %s with %s", filePath, str(e))
            raise e

        return False

    @retry((requests.exceptions.RequestException), maxAttempts=3, delaySeconds=5, multiplier=3, defaultValue=False, logger=logger)
    def __fetchUrlReqRt(self, url, filePath, **kwargs):
        return self.__fetchUrlReqNoRt(url, filePath, **kwargs)

    def __fetchUrlReqNoRt(self, url, filePath, **kwargs):
        """Fetch data from a remote URL and store this in input filePath.

        Args:
            url (str): target URL to fetch
            filePath (str): path to store the data from the remote url
            username (str): basic auth username
            password (str): dasic auth password
            kwargs (dict, optional): other options

        Raises:
            e: any exception

        Returns:
            bool: True for sucess or False otherwise

        """
        user = kwargs.get("username", None)
        pw = kwargs.get("password", None)
        chunkSize = kwargs.get("chunkSize", 1024 * 1024 * 3)
        #
        try:
            ok = self.mkdirForFile(filePath)
            if not ok:
                logger.error("Failing to create target directory for file %r", filePath)
                return False
            if user and pw:
                with requests.get(url, stream=True, auth=(user, pw), allow_redirects=True) as rIn:
                    if rIn.status_code == requests.codes.ok:  # pylint: disable=no-member
                        with open(filePath, "wb") as fOut:
                            for chunk in rIn.iter_content(chunk_size=chunkSize):
                                fOut.write(chunk)
                    else:
                        logger.error("Fetch %r fails with status %r", url, rIn.status_code)
                        rIn.raise_for_status()
                        return False
            else:
                with requests.get(url, stream=True, allow_redirects=True) as rIn:
                    if rIn.status_code == requests.codes.ok:  # pylint: disable=no-member
                        with open(filePath, "wb") as fOut:
                            for chunk in rIn.iter_content(chunk_size=chunkSize):
                                fOut.write(chunk)
                    else:
                        logger.error("Fetch %r fails with status %r", url, rIn.status_code)
                        rIn.raise_for_status()
                        return False

            return True
        except Exception as e:
            logger.error("Failing for %s with %s", filePath, str(e))
            raise e

        return False

    def toAscii(self, inputFilePath, outputFilePath, chunkSize=5000, encodingErrors="ignore"):
        """Encode input file to Ascii and write this to the target output file.   Handle encoding
        errors according to the input settting ('ignore', 'escape', 'xmlcharrefreplace').
        """
        try:
            ok = self.mkdirForFile(outputFilePath)
            if not ok:
                logger.error("Failing to create target directory for file %r", outputFilePath)
                return False
            chunk = []
            with io.open(inputFilePath, "r", encoding="utf-8") as rIn, io.open(outputFilePath, "w", encoding="ascii") as wOut:
                for line in rIn:
                    # chunk.append(line.encode('ascii', 'xmlcharrefreplace').decode('ascii'))
                    chunk.append(line.encode("ascii", encodingErrors).decode("ascii"))
                    if len(chunk) == chunkSize:
                        wOut.writelines(chunk)
                        chunk = []
                wOut.writelines(chunk)
            return True
        except Exception as e:
            logger.error("Failing text ascii encoding for %s with %s", inputFilePath, str(e))
        #
        return False

    def uncompress(self, inputFilePath, outputDir=None):
        """Uncompress the input file if the path name has a recognized compression type file extension.

        Return the path of the uncompressed file (in outDir) or the original input file path.

        """
        try:
            outputDir = outputDir if outputDir else tempfile.gettempdir()
            _, fn = os.path.split(inputFilePath)
            bn, _ = os.path.splitext(fn)
            outputFilePath = os.path.join(outputDir, bn)
            if inputFilePath.endswith(".gz"):
                with gzip.open(inputFilePath, mode="rb") as inpF:
                    with io.open(outputFilePath, "wb") as outF:
                        shutil.copyfileobj(inpF, outF)
            elif inputFilePath.endswith(".bz2"):
                with bz2.open(inputFilePath, mode="rb") as inpF:
                    with io.open(outputFilePath, "wb") as outF:
                        shutil.copyfileobj(inpF, outF)
            # elif inputFilePath.endswith(".xz"):
            #    with lzma.open(inputFilePath, mode="rb") as inpF:
            #        with io.open(outputFilePath, "wb") as outF:
            #            shutil.copyfileobj(inpF, outF)
            elif inputFilePath.endswith(".zip"):
                with zipfile.ZipFile(inputFilePath, mode="r") as zObj:
                    memberList = zObj.namelist()
                    for member in memberList[:1]:
                        zObj.extract(member, path=outputDir)
                if memberList:
                    outputFilePath = os.path.join(outputDir, memberList[0])
            else:
                outputFilePath = inputFilePath

        except Exception as e:
            logger.exception("Failing uncompress for file %s with %s", inputFilePath, str(e))
        logger.debug("Returning file path %r", outputFilePath)
        return outputFilePath

    def compress(self, inpPath, outPath, compressType="gzip"):
        try:
            if compressType == "gzip":
                with open(inpPath, "rb") as fIn:
                    with gzip.open(outPath, "wb") as fOut:
                        shutil.copyfileobj(fIn, fOut)
                return True
            else:
                logger.error("Unsupported compressType %r", compressType)
        except Exception as e:
            logger.exception("Compressing file %s failing with %s", inpPath, str(e))
        return False



##
# File:    SftpUtil.py
# Author:  jdw
# Date:    5-Jun-2020
#
# Updates:
#
##
"""
Class providing essential data transfer operations for SFTP protocol.
"""

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"

#
import logging

import paramiko

from rcsb.utils.io.FileUtil import FileUtil

logger = logging.getLogger(__name__)


class SftpUtil(object):
    """Class providing essential data transfer operations for SFTP protocol"""

    def __init__(self, *args, **kwargs):
        _ = args
        _ = kwargs
        self.__sftpClient = None
        self.__transport = None
        self.__raiseExceptions = False
        #

    # def getRootPath(self):
    #    return self._rootPath

    def connect(self, hostName, userName, pw=None, port=22, keyFilePath=None, keyFileType="RSA"):

        try:
            self.__sftpClient = self.__makeSftpClient(hostName=hostName, port=port, userName=userName, pw=pw, keyFilePath=keyFilePath, keyFileType=keyFileType)
            return True
        except Exception as e:
            if self.__raiseExceptions:
                raise e
            else:
                logger.error("Failing connect for hostname %s with %s", hostName, str(e))
                return False

    def __makeSftpClient(self, hostName, port, userName, pw=None, keyFilePath=None, keyFileType="RSA"):
        """
        Make SFTP client connected to the supplied host on the supplied port authenticating as the user with
        supplied username and supplied password or with the private key in a file with the supplied path.
        If a private key is used for authentication, the type of the keyfile needs to be specified as DSA or RSA.

        :rtype: Paramiko SFTPClient object.

        """
        sftp = None
        key = None
        self.__transport = None
        try:
            if keyFilePath is not None:
                # Get private key used to authenticate user.
                if keyFileType == "DSA":
                    # The private key is a DSA type key.
                    key = paramiko.DSSKey.from_private_key_file(keyFilePath)
                else:
                    # The private key is a RSA type key.
                    key = paramiko.RSAKey.from_private_key_file(keyFilePath)

            # Create Transport object using supplied method of authentication.
            self.__transport = paramiko.Transport((hostName, port))
            if pw is not None:
                self.__transport.connect(username=userName, password=pw)
            else:
                self.__transport.connect(username=userName, pkey=key)

            sftp = paramiko.SFTPClient.from_transport(self.__transport)

            return sftp
        except Exception as e:
            logger.error("Error occurred creating SFTP client: %s: %s", e.__class__, str(e))
            if sftp is not None:
                sftp.close()
            if self.__transport is not None:
                self.__transport.close()
            if self.__raiseExceptions:
                raise e

    def mkdir(self, path, mode=511):

        try:
            self.__sftpClient.mkdir(path, mode)
            return True
        except Exception as e:
            if self.__raiseExceptions:
                raise e
            else:
                logger.error("mkdir failing for path %s with %s", path, str(e))
                return False

    def stat(self, path):
        """sftp  stat attributes  = [ size=17 uid=0 gid=0 mode=040755 atime=1507723473 mtime=1506956503 ]"""
        try:
            sT = self.__sftpClient.stat(path)
            dD = {"mtime": sT.st_mtime, "size": sT.st_size, "mode": sT.st_mode, "uid": sT.st_uid, "gid": sT.st_gid, "atime": sT.st_atime}
            return dD
        except Exception as e:
            if self.__raiseExceptions:
                raise e
            else:
                logger.error("stat failing for path %s with %s", path, str(e))
                return {}

    def put(self, localPath, remotePath):
        try:
            self.__sftpClient.put(localPath, remotePath)
            return True
        except Exception as e:
            if self.__raiseExceptions:
                raise e
            else:
                logger.error("put failing for localPath %s  remotePath %s with %s", localPath, remotePath, str(e))
                return False

    def get(self, remotePath, localPath):
        try:
            fileU = FileUtil()
            fileU.mkdirForFile(localPath)
            self.__sftpClient.get(remotePath, localPath)
            return True
        except Exception as e:
            if self.__raiseExceptions:
                raise e
            else:
                logger.error("get failing for remotePath %s localPath %s with %s", remotePath, localPath, str(e))
                return False

    def listdir(self, path):
        try:
            return self.__sftpClient.listdir(path)
        except Exception as e:
            if self.__raiseExceptions:
                raise e
            else:
                logger.error("listdir failing for path %s with %s", path, str(e))
                return False

    def rmdir(self, dirPath):
        try:
            self.__sftpClient.rmdir(dirPath)
            return True
        except Exception as e:
            if self.__raiseExceptions:
                raise e
            else:
                logger.error("rmdir failing for path %s with %s", dirPath, str(e))
                return False

    def remove(self, filePath):
        try:
            self.__sftpClient.remove(filePath)
            return True
        except Exception as e:
            if self.__raiseExceptions:
                raise e
            else:
                logger.error("remove failing for path %s with %s", filePath, str(e))
                return False

    def close(self):
        try:
            if self.__sftpClient is not None:
                self.__sftpClient.close()
            if self.__transport is not None:
                self.__transport.close()

            return True
        except Exception as e:
            if self.__raiseExceptions:
                raise e
            else:
                logger.error("Close failing with %s", str(e))



##
# File:    ProcessStatusUtil.py
# Author:  J. Westbrook
# Date:    23-Aug-2020
# Version: 0.001
#
# Updates:
#
##
"""
Convenience utilities to report process status.

"""

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import datetime
import logging
import platform
import socket
import time
import uuid

# import requests

import psutil

logger = logging.getLogger(__name__)


class ProcessStatusUtil(object):
    def __init__(self, **kwargs):
        pass

    def getInfo(self):
        """Return a dictionary of system and process status details.

        Returns:
            dict: dictionary of system and process status
        """
        infoD = {}
        infoD.update(self.__getSystemInfo())
        infoD["traffic"] = self.__getNetworkTraffic()
        infoD.update(self.__getCpuInfo())
        infoD["memory"] = self.__getMemoryInfo()
        infoD.update(self.__getStorageInfo())
        #
        return infoD

    def __getSystemInfo(self):
        rD = {}
        try:
            hostName = socket.gethostname()
            systemName = platform.system()
            systemVersion = platform.release()
            timeStamp = datetime.datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%S+00:00")
            uptimeSeconds = int(time.time() - psutil.boot_time())
            systemId = uuid.getnode()
            rD = {"hostName": hostName, "systemName": systemName, "systemVersion": systemVersion, "update": timeStamp, "uptimeSeconds": uptimeSeconds, "uuid": systemId}
        except Exception as e:
            logger.exception("Failing with %r", str(e))
        return rD

    def __getCpuInfo(self):
        rD = {}
        try:
            cpuCount = psutil.cpu_count()
            cpuUsage = psutil.cpu_percent(interval=1)
            rD = {"cpuCount": cpuCount, "cpuUsagePercent": cpuUsage}
        except Exception as e:
            logger.exception("Failing with %r", str(e))
        return rD

    def __getMemoryInfo(self):
        rD = {}
        try:
            vm = psutil.virtual_memory()
            memoryTotalGi = round(vm.total / 1e9, 2)
            memoryUsedGi = round(vm.used / 1e9, 2)
            memoryUsedPercent = vm.percent
            rD = {"memoryTotalGi": memoryTotalGi, "memoryUsedGi": memoryUsedGi, "memoryUsedPercent": memoryUsedPercent}
        except Exception as e:
            logger.exception("Failing with %r", str(e))
        return rD

    def __getStorageInfo(self):
        rD = {}
        try:
            diskPartL = psutil.disk_partitions()
            diskL = []
            for dp in diskPartL:
                try:
                    disk = {
                        "name": dp.device,
                        "mount_point": dp.mountpoint,
                        "type": dp.fstype,
                        "totalSizeGi": round(psutil.disk_usage(dp.mountpoint).total / 1e9, 1),
                        "usedSizeGi": round(psutil.disk_usage(dp.mountpoint).used / 1e9, 1),
                        "percentUsed": psutil.disk_usage(dp.mountpoint).percent,
                    }
                    diskL.append(disk)
                except Exception:
                    pass
            rD = {"diskInfo": diskL}
        except Exception as e:
            logger.exception("Failing with %r", str(e))
        return rD

    def __getNetworkInfo(self):
        rD = {"networkInterfaces": []}
        try:
            # Network Info
            nicL = []
            for name, snicL in psutil.net_if_addrs().items():
                # Create NIC object
                nic = {"name": name, "mac": "", "address": "", "address6": "", "netmask": ""}
                # Get NiC values
                for snic in snicL:
                    if snic.family == -1:
                        nic["mac"] = snic.address
                    elif snic.family == 2:
                        nic["address"] = snic.address
                        nic["netmask"] = snic.netmask
                    elif snic.family == 23:
                        nic["address6"] = snic.address
                nicL.append(nic)

            rD = {"networkInterfaces": nicL}
        except Exception as e:
            logger.exception("Failing with %r", str(e))
        return rD

    def __getNetworkTraffic(self):
        rD = {"trafficIn": 0, "trafficOut": 0}
        try:
            #
            net1Out = psutil.net_io_counters().bytes_sent
            net1In = psutil.net_io_counters().bytes_recv
            time.sleep(1)
            net2Out = psutil.net_io_counters().bytes_sent
            net2In = psutil.net_io_counters().bytes_recv

            # Compare and get current speed
            if net1In > net2In:
                currentIn = 0
            else:
                currentIn = net2In - net1In

            if net1Out > net2Out:
                currentOut = 0
            else:
                currentOut = net2Out - net1Out
            rD = {"trafficIn": currentIn, "trafficOut": currentOut}
        except Exception as e:
            logger.exception("Failing with %r", str(e))
        return rD



##
# File: CryptUtils.py
# Date: 12-Aug-2019 Jdw
#
# Updates:
#
##
"""
 Utilities supporting encrypting and decrypting flat files and messages.
"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"

import base64
import binascii
import hashlib
import logging

from nacl.encoding import HexEncoder, RawEncoder
from nacl.secret import SecretBox
from nacl.utils import random

logger = logging.getLogger(__name__)


class CryptUtils(object):
    """   Utility class for encrypting and decrypting data files and messages

          ** This class does not yet support any stream operations so resources must be
          available to load data sets in memory -

    """

    def __init__(self, *args, **kwargs):
        _ = args
        _ = kwargs

    def newKey(self):
        """Generate a new encryption key.

        Returns:
            (str): encryption key
        """
        key = random(SecretBox.KEY_SIZE)
        return binascii.hexlify(key).decode("utf-8")

    def encryptMessage(self, msg, hexKey):
        """Encrypt the input message.

        Args:
            msg (str): input message
            hexKey (str): encryption key

        Returns:
            (str):  encrypted message text
        """
        txt = None
        try:
            box = SecretBox(hexKey.encode("utf-8"), encoder=HexEncoder)
            bMsg = msg.encode("utf-8")
            encMsg = box.encrypt(bMsg, encoder=RawEncoder)
            txt = base64.b64encode(encMsg).decode("ascii")
        except Exception as e:
            logger.exception("Failing with %s", str(e))

        return txt

    def decryptMessage(self, msg, hexKey):
        """Decrypt the input message.

        Args:
            msg (str): input message
            hexKey (str): encryption key

        Returns:
            (str):  decrypted message text
        """
        txt = None
        try:
            box = SecretBox(hexKey.encode("utf-8"), encoder=HexEncoder)
            bMsg = base64.b64decode(msg)
            dcrMsg = box.decrypt(bMsg)
            logger.debug("type %r text %r", type(dcrMsg), dcrMsg)
            txt = dcrMsg.decode("utf-8")
        except Exception as e:
            logger.exception("Failing with %s", str(e))

        return txt

    def encryptFile(self, inpPath, outPath, hexKey, chunkSize=65536):
        """Encrypt the data in the input file store this in the output file.

        Args:
            inpPath (str): input file path
            outPath (str): output file path (encrypted)
            hexKey (str): encryption key
            chunkSize (int, optional): the size of incremental copy operations. Defaults to 65536.

        Returns:
            (bool): True for success or False otherwise
        """
        try:
            box = SecretBox(hexKey.encode("utf-8"), encoder=HexEncoder)
            tL = []
            with open(inpPath, "rb") as ifh:
                while True:
                    chunk = ifh.read(chunkSize)
                    if not chunk:
                        break
                    tL.append(chunk)
            #
            txt = b"".join(tL)
            logger.debug("Read %s size %d", inpPath, len(txt))
            enctxt = box.encrypt(txt, encoder=RawEncoder)
            tL = [enctxt[i : i + chunkSize] for i in range(0, len(enctxt), chunkSize)]
            logger.debug("Chunks %d size %d", len(tL), len(enctxt))
            #
            with open(outPath, "wb") as ofh:
                for tV in tL:
                    ofh.write(tV)
            return True
        except Exception as e:
            logger.exception("Failing with %r", str(e))
        return False

    def decryptFile(self, inpPath, outPath, hexKey, chunkSize=65536):
        """Decrypt the data in the input file store the result in the output file.

        Args:
            inpPath ([type]): input file path (encrypted)
            outPath ([type]): output file path
            hexKey ([type]): encryption key
            chunkSize (int, optional): the size of incremental copy operations. Defaults to 65536.

        Returns:
            (bool): True for success or False otherwise
        """
        try:
            box = SecretBox(hexKey.encode("utf-8"), encoder=HexEncoder)
            tL = []
            with open(inpPath, "rb") as ifh:
                while True:
                    chunk = ifh.read(chunkSize)
                    if not chunk:
                        break
                    tL.append(chunk)
            #
            enctxt = b"".join(tL)
            #
            txt = box.decrypt(enctxt)
            tL = [txt[i : i + chunkSize] for i in range(0, len(txt), chunkSize)]
            logger.debug("Chunks %d size %d", len(tL), len(txt))
            #
            with open(outPath, "wb") as ofh:
                for tV in tL:
                    ofh.write(tV)
            return True
        except Exception as e:
            logger.exception("Failing with %r", str(e))
        return False

    def getFileHash(self, filePath, hashType="SHA1", blockSize=65536):
        """Return the hash (hashType) for the input file.

        Args:
            filePath (str): for input file path
            hashType (str, optional): one of MD5, SHA1, or SHA256. Defaults to "SHA1".
            blockSize (int, optional): the size of incremental read operations. Defaults to 65536.

        Returns:
            (dict): {"hashDigest": xxxx , "hashType": 'SHA1|MD5|SHA256'}
        """
        rD = {}
        if hashType not in ["MD5", "SHA1", "SHA256"]:
            return rD
        try:
            if hashType == "SHA1":
                hashObj = hashlib.sha1()
            elif hashType == "SHA256":
                hashObj = hashlib.sha256()
            elif hashType == "MD5":
                hashObj = hashlib.md5()

            with open(filePath, "rb") as ifh:
                for block in iter(lambda: ifh.read(blockSize), b""):
                    hashObj.update(block)
            rD = {"hashDigest": hashObj.hexdigest(), "hashType": hashType}
        except Exception as e:
            logger.exception("Failing with file %s %r", filePath, str(e))
        return rD
